/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#include "libnnfc.h"

int n_socket(int domain, int type, int protocol)
{
    int rc;

    rc = socket(domain, type, protocol);
    if (rc < 0) {
        syslog(LOG_CRIT, "socket: %s", strerror(errno));
        closelog();
        exit(-1);
    }
    return rc;
}

void n_bind(int sockfd, struct sockaddr *my_addr, socklen_t addrlen)
{
    int rc;
	
    rc = bind(sockfd, my_addr, addrlen);
    if (rc < 0) {
        syslog(LOG_CRIT, "bind: %s", strerror(errno));
        closelog();
        if (sockfd > 0)
            n_shutdown(sockfd, SHUT_RDWR);
        exit(-1);
    }
    return;
}

ssize_t n_recvfrom(int s, void *buf, size_t len, int flags, 
        struct sockaddr *from, socklen_t *fromlen)
{
    ssize_t rc;

    rc = recvfrom(s, buf, len, flags, from, fromlen);
    if (rc < 0 || rc > len) {
        syslog(LOG_ERR, "recvfrom: %s", strerror(errno));
        return 0;
    } else {
        return rc;
    }
}

int n_select(int n, fd_set *readfds, fd_set *writefds, fd_set *exceptfds,
        struct timeval *timeout)
{
    int rc;

    rc = select(n, readfds, writefds, exceptfds, timeout);
    if (rc < 0) {
        syslog(LOG_ERR, "select: %s", strerror(errno));
    }
    return rc;
}

void n_shutdown(int s, int how)
{
    int rc;

    if (s <= 0) return;
    rc = shutdown(s, how);
    if (rc < 0) {
        syslog(LOG_ERR, "shutdown: %s", strerror(errno));
        close(s);
    }
    return;
}

void n_daemon()
{
    int rc;

    rc = daemon(1, 1);
    if (rc < 0) {
        syslog(LOG_ERR, "daemon: %s", strerror(errno));
    }
    return;
}

/* sock_ntop function is from:
 * W. Richard Stevens. UNIX Network Programming.
 */
 
char *sock_ntop(const struct sockaddr *sa, socklen_t salen)
{
    char        portstr[7];
    static char str[128];		/* Unix domain is largest */

    switch (sa->sa_family) {
    case AF_INET: {
        struct sockaddr_in	*sin = (struct sockaddr_in *) sa;
        if (inet_ntop(AF_INET, &sin->sin_addr, str, sizeof(str)) == NULL)
            return(NULL);
        if (ntohs(sin->sin_port) != 0) {
            snprintf(portstr, sizeof(portstr), ".%d", ntohs(sin->sin_port));
            strncat(str, portstr, 6);
        }
        return(str);
    }
    /* end sock_ntop */
    default:
        snprintf(str, sizeof(str), "sock_ntop: unknown AF_xxx: %d, len %d", 
            sa->sa_family, salen);
        return(str);
    }
    return (NULL);
}

char *n_sock_ntop(const struct sockaddr *sa, socklen_t salen)
{
    char    *ptr;

    if ( (ptr = sock_ntop(sa, salen)) == NULL)
        syslog(LOG_ERR, "sock_ntop: %s", strerror(errno));
    return(ptr); 
}

