/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>

#include <libnnfc.h>

/*
 * Parameters for this plugins are:
 * localtime - store local or remote time?
 * hdrfile - header file template
 * flowfile - flow file template
 */

static FILE	*fhdr;
static FILE	*fflow;

int     ltime = 1;
char    *hdrtmpl = 0;
char    *flwtmpl = 0;

void * _create_thread_context()
{
}

/* creating file */
void _connect(char *fname, void *param)
{
    struct command  *commands = 0;
    int             n = 0;

    if (!fname) return;

    if (readcfg(fname, &commands)) {
        syslog(LOG_ERR, "Parse error!\n");
        return;
    }

    CFG_GETI(ltime, "localtime", commands);
    CFG_GETS(hdrtmpl, "hdrfile", commands);
    CFG_GETS(flwtmpl, "flowfile", commands);

    freecmds(commands);
}

/* closes connection to database */
void _close(void *param)
{
    if (fhdr) fclose(fhdr);
    if (fflow) fclose(fflow);

    if (hdrtmpl) free(hdrtmpl);
    if (flwtmpl) free(flwtmpl);
}

void begin(void *param)
{
    char        fhname[MAXPATHLEN];
    char        fnname[MAXPATHLEN];
    time_t      ttm;
    struct tm   *stm;

    ttm = time(NULL);
    stm = localtime(&ttm);
    strftime(fhname, MAXPATHLEN, hdrtmpl, stm);
    strftime(fnname, MAXPATHLEN, flwtmpl, stm);
    fhdr = fopen(fhname, "a+");
    if (!fhdr) {
        syslog(LOG_ERR, "Cannot open archive file for writing!");
        return;
    }
    fflow = fopen(fnname, "a+");
    if (!fflow) {
        syslog(LOG_ERR, "Cannot open archive file for writing!");
        return;
    }

    return;
}

void commit(void *param)
{
    if (fhdr) {
        fflush(fhdr);
        fclose(fhdr);
    }
    if (fflow) {
        fflush(fflow);
        fclose(fflow);
    }
    return;
}

/* inserts netflow packet's header into database */
void inserthdr(uint32_t address, void *flowhdr, void *param)
{
    struct in_addr  in;
    char            str[128];
    int             version = ntohs(*((short *)flowhdr));

    if (!fhdr) 
        return;

    if (!flowhdr) 
        return;

    in.s_addr = address;
    if (!inet_ntop(AF_INET, &in, str, sizeof(str)))
        strncpy(str, "0.0.0.0", 7);

    switch (version) {
    case 1: {           /* netflow v1 */
        struct flowhdr1_t   *hdr = (struct flowhdr1_t *)flowhdr;
        fprintf(fhdr, "'%s' %u %u %lu %lu %lu\n", str, htons(hdr->version),
            htons(hdr->count), (unsigned long)htonl(hdr->uptime),
            (unsigned long)htonl(hdr->secs), (unsigned long)htonl(hdr->nsecs));
        break;
    }
    case 5: {           /* netflow v5 */
        struct flowhdr5_t   *hdr = (struct flowhdr5_t *)flowhdr;
        fprintf(fhdr, "'%s' %u %u %lu %lu %lu %lu %u %u %u\n", str,
            htons(hdr->version), htons(hdr->count),
            (unsigned long)htonl(hdr->uptime), (unsigned long)htonl(hdr->secs),
            (unsigned long)htonl(hdr->nsecs),
            (unsigned long)htonl(hdr->flowseq), hdr->engine, hdr->eng_id,
            htons(hdr->sampling));
        break;
    }
    case 7: {           /* netflow v7 */
        struct flowhdr7_t   *hdr = (struct flowhdr7_t *)flowhdr;
        fprintf(fhdr, "'%s' %u %u %lu %lu %lu %lu %lu\n", str,
            htons(hdr->version), htons(hdr->count),
            (unsigned long)htonl(hdr->uptime), (unsigned long)htonl(hdr->secs),
            (unsigned long)htonl(hdr->nsecs),
            (unsigned long)htonl(hdr->flowseq),
            (unsigned long)htonl(hdr->resv));
    }
    }
    fflush(fhdr);
}

/* inserts flow into database */
void insertflow(uint32_t address, void *flowhdr, void *flow, int num, void *param)
{
    struct in_addr  in;
    char            str[128];
    int             version = ntohs(*((short *)flowhdr));
    struct tm       *stm;
    time_t          ttm;
    unsigned int    uptime;
    unsigned int    luptime;

    struct flowhdr5_t   *hdr5 = (struct flowhdr5_t *)flowhdr;
    struct flow1_t      *flow1 = (struct flow1_t *)flow;
    struct flow5_t      *flow5 = (struct flow5_t *)flow;
    struct flow7_t      *flow7 = (struct flow7_t *)flow;

    if (!fflow) 
        return;

    if (!flowhdr || !flow) 
        return;

    in.s_addr = address;
    if (!inet_ntop(AF_INET, &in, str, sizeof(str))) 
        strncpy(str, "0.0.0.0", 7);

    if (ltime)
        ttm = time(NULL);
    else {
        uptime = htonl((unsigned int)hdr5->uptime);
        luptime = htonl((unsigned int)flow5->luptime);

        if (uptime > luptime) {
            ttm = htonl(hdr5->secs) - (uptime - luptime)/1000;
        } else {
            ttm = htonl(hdr5->secs) + (luptime - uptime)/1000;
        }
    }
    stm = localtime(&ttm);

    switch (version) {
    case 1:             /* netflow v1 */
        fprintf(fflow, "'%s' %u %u '%u.%u.%u.%u' '%u.%u.%u.%u'"
            " '%u.%u.%u.%u' %u %u %lu %lu %lu %lu %u %u %u"
            " %u %u %u %u %u %u %lu '%04d-%02d-%02d %02d:%02d:%02d'\n",
            str, version, num, flow1->srcaddr[0], flow1->srcaddr[1],
            flow1->srcaddr[2], flow1->srcaddr[3],
            flow1->dstaddr[0], flow1->dstaddr[1],
            flow1->dstaddr[2], flow1->dstaddr[3],
            flow1->nexthop[0], flow1->nexthop[1],
            flow1->nexthop[2], flow1->nexthop[3],
            htons(flow1->input_if), htons(flow1->output_if),
            (unsigned long)htonl(flow1->dpkts),
            (unsigned long)htonl(flow1->doctets),
            (unsigned long)htonl(flow1->fuptime),
            (unsigned long)htonl(flow1->luptime),
            htons(flow1->srcport), htons(flow1->dstport), htons(flow1->resv1),
            flow1->prot, flow1->tos, flow1->flags, flow1->tcp_retx_cnt,
            flow1->tcp_retx_sec, flow1->tcp_misseq_cnt,
            (unsigned long)htonl(flow1->resv2),
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec
            );
        break;
	case 5:             /* netflow v5 */
        fprintf(fflow, "'%s' %u %lu '%u.%u.%u.%u' '%u.%u.%u.%u'"
            " '%u.%u.%u.%u' %u %u %lu %lu %lu %lu %u %u %u"
            " %u %u %u %u %u %u %u %u '%04d-%02d-%02d %02d:%02d:%02d'\n",
            str, version, (unsigned long)htonl(hdr5->flowseq)+num,
            flow5->srcaddr[0], flow5->srcaddr[1],
            flow5->srcaddr[2], flow5->srcaddr[3],
            flow5->dstaddr[0], flow5->dstaddr[1],
            flow5->dstaddr[2], flow5->dstaddr[3],
            flow5->nexthop[0], flow5->nexthop[1],
            flow5->nexthop[2], flow5->nexthop[3],
            htons(flow5->input_if), htons(flow5->output_if),
            (unsigned long)htonl(flow5->dpkts),
            (unsigned long)htonl(flow5->doctets),
            (unsigned long)htonl(flow5->fuptime),
            (unsigned long)htonl(flow5->luptime),
            htons(flow5->srcport), htons(flow5->dstport),
            (unsigned char)flow5->resv1, flow5->flags, flow5->prot, flow5->tos,
            htons(flow5->src_as), htons(flow5->dst_as), flow5->srcmask,
            flow5->dstmask, htons(flow5->resv2),
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec
            );
        break;
	case 7:             /* netflow v7 */
        fprintf(fflow, "'%s' %u %lu '%u.%u.%u.%u' '%u.%u.%u.%u' "
            "'%u.%u.%u.%u' %u %u %lu %lu %lu %lu %u %u %u "
            "%u %u %u %u %u %u %u %u '%u.%u.%u.%u' "
            "'%04d-%02d-%02d %02d:%02d:%02d'\n",
            str, version, (unsigned long)htonl(hdr5->flowseq)+num,
            flow7->srcaddr[0], flow7->srcaddr[1],
            flow7->srcaddr[2], flow7->srcaddr[3],
            flow7->dstaddr[0], flow7->dstaddr[1],
            flow7->dstaddr[2], flow7->dstaddr[3],
            flow7->nexthop[0], flow7->nexthop[1],
            flow7->nexthop[2], flow7->nexthop[3],
            htons(flow7->input_if), htons(flow7->output_if),
            (unsigned long)htonl(flow7->dpkts),
            (unsigned long)htonl(flow7->doctets),
            (unsigned long)htonl(flow7->fuptime),
            (unsigned long)htonl(flow7->luptime),
            htons(flow7->srcport), htons(flow7->dstport),
            (unsigned char)flow7->mode, (unsigned char)flow7->flags,
            (unsigned char)flow7->prot, (unsigned char)flow7->tos,
            htons(flow7->src_as), htons(flow7->dst_as),
            (unsigned char)flow7->srcmask, (unsigned char)flow7->dstmask,
            htons(flow7->resv2), flow7->router[0], flow7->router[1],
            flow7->router[2], flow7->router[3],
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec
            );
        break;
    }
    fflush(fflow);
}

void insertpkt(uint32_t address, void *packet, void *param)
{
    struct flowhdr1_t   *hdr1 = (struct flowhdr1_t *)packet;
    short               version = ntohs(*((short *)packet));
    int                 i;
    void                *flow;

    if (!packet) return;

    inserthdr(address, packet, param);

    for (i = 0; i < htons(hdr1->count); i++) {
        switch (version) {
        case 1:
            flow = packet + sizeof(struct flowhdr1_t) +
                sizeof(struct flow1_t)*i;
            break;
        case 5:
            flow = packet + sizeof(struct flowhdr5_t) +
                sizeof(struct flow5_t)*i;
            break;
        case 7:
            flow = packet + sizeof(struct flowhdr7_t) +
                sizeof(struct flow7_t)*i;
            break;
        }
        insertflow(address, packet, flow, i, param);
    }
}

