/**************************************************************************
 * Copyright (c) 2004-2006, Andrew Klyachkin <andrewk@region73.ru>        *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Andrew Klyachkin nor the names of his       *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <libpq-fe.h>
#include <netdb.h>
#include <errno.h>

#include <libnnfc.h>

static PGconn   *conn = 0;
char            *config = 0;
static int      transaction = 0;

/*
 *  Parameters:
 *
 *  localtime - store local or remote time
 *  savehdr - store flow headers
 *  dbhost - postgres server
 *  dbport - postgres server's port
 *  dbuser - postgres user's name
 *  dbpass - postgres user's password
 *  dbname - postgres database
 *
 *  dbheader - postgres table for netflow headers
 *  dbflows - postgres table for internal local traffic
 *  dbextflows - postgres table for external internet traffic
 *  dbproxy - postgres table for proxy server's traffic
 *
 *  ip - list of our routers' IP addresses
 *  proxy - proxy server's IP address
 *  proxyport - proxy server's port
 *  border - IP address of border router
 *  intf - SNMP indices of external interfaces in border router
 */

int         ltime = 1;
int         dbport = 5432;
int         proxyport = 3128;
int         savehdr = 0;
char        *dbhost = 0;
char        *dbuser = 0;
char        *dbpass = 0;
char        *dbname = 0;
char        *dbheader = 0;
char        *dbflows = 0;
char        *dbextflows = 0;
char        *dbproxy = 0;
uint32_t    *ip_addr = 0;
int         addr_count = 0;
uint32_t    proxy = 0;
uint32_t    border = 0;
int         *intf = 0;
int         intf_count = 0;

void _close(void *);
void _connect(char *, void *);
void begin(void *);
void commit(void *);
void insertpkt(uint32_t, void *, void *);

void * _create_thread_context()
{
}

void _connect(char *fname, void *param)
{
    char            str[128];
    struct command  *commands = 0;
    int             n = 0;
    struct in_addr  addr;
    struct hostent  *he;
    struct value    *v = 0;

    if (!fname) return;

    if (!config) {
        config = (char *)calloc(1, MAXPATHLEN+1);
        if (strlen(fname) > MAXPATHLEN)
            strncpy(config, fname, MAXPATHLEN);
        else
            strncpy(config, fname, strlen(fname));
    }

    if (readcfg(config, &commands)) {
        syslog(LOG_ERR, "megabit: parse error!");
        return;
    }

    CFG_GETI(ltime, "localtime", commands);
    CFG_GETI(dbport, "dbport", commands);
    CFG_GETI(proxyport, "proxyport", commands);
    CFG_GETI(savehdr, "savehdr", commands);

    CFG_GETS(dbhost, "dbhost", commands);
    CFG_GETS(dbuser, "dbuser", commands);
    CFG_GETS(dbpass, "dbpass", commands);
    CFG_GETS(dbname, "dbname", commands);
    CFG_GETS(dbheader, "dbheader", commands);
    CFG_GETS(dbflows, "dbflows", commands);
    CFG_GETS(dbextflows, "dbextflows", commands);
    CFG_GETS(dbproxy, "dbproxy", commands);

    /* proxy IP address */
    if (getvalue("proxy", commands)) {
        if (atoi(getvalue("proxy", commands)) == 0)
            proxy = 0;
        else {
            if (inet_pton(AF_INET, getvalue("proxy", commands), &addr) <= 0) {
                if ((he = gethostbyname(getvalue("proxy", commands))) == 0) {
                    syslog(LOG_ERR, "megabit: Invalid address in proxy field");
                    return;
                }
                addr = *(struct in_addr *)(he->h_addr);
            }
            proxy = addr.s_addr;
        }
    }

    /* border router's IP address */
    if (getvalue("border", commands)) {
        if (atoi(getvalue("border", commands)) == 0)
            border = 0;
        else {
            if (inet_pton(AF_INET, getvalue("border", commands), &addr) <= 0) {
                if ((he = gethostbyname(getvalue("border", commands))) == 0) {
                    syslog(LOG_ERR, "megabit: Invalid address in border field");
                    return;
                }
                addr = *(struct in_addr *)(he->h_addr);
            }
            border = addr.s_addr;
        }
    }

    /* routers */
    v = getallvalues("ip", commands);
    addr_count = 0;
    while (v) {
        if (inet_pton(AF_INET, v->value, &addr) <= 0) {
            if ((he = gethostbyname(v->value)) == 0) {
                syslog(LOG_ERR, "megabit: invalid address in parameter ip");
                return;
            }
            addr = *(struct in_addr *)(he->h_addr);
        }
        if (!ip_addr) {
            ip_addr = (uint32_t *)calloc(1, sizeof(uint32_t));
        } else {
            ip_addr = (uint32_t *)realloc(ip_addr,
                sizeof(uint32_t)*(addr_count+1));
        }
        ip_addr[addr_count] = addr.s_addr;
        addr_count++;
        v = v->next;
    }

    /* interfaces */
    v = getallvalues("intf", commands);
    intf_count = 0;
    while (v) {
        if (!intf) {
            intf = (int *)calloc(1, sizeof(int));
        } else {
            intf = (int *)realloc(intf, sizeof(int)*(intf_count+1));
        }
        errno = 0;
        intf[intf_count] = strtol(v->value, (char **)NULL, 10);
        if (errno) {
            syslog(LOG_ERR, "megabit: not a number in intf field");
            return;
        }
        intf_count++;
        v = v->next;
    }

    freecmds(commands);

    if (!dbport || !dbhost || !dbuser || !dbpass || !dbname ||
            !dbflows || !dbextflows || !dbproxy || !proxyport || 
            !ip_addr || !intf || !proxy || !border) {

        syslog(LOG_ERR, "megabit: one of the required parameters is missing");
        return;
    }

    memset((void *)&str, 0, 128);
    snprintf(str, sizeof(str), "host=%s port=%d dbname=%s user=%s password=%s",
        dbhost, dbport, dbname, dbuser, dbpass);
    conn = PQconnectdb(str);
    if (!conn || PQstatus(conn) == CONNECTION_BAD) {
        conn = 0;
        syslog(LOG_ERR, "megabit: cannot establish connection to %s\n", dbhost);
        return;
    }
}

void begin(void *param)
{
    PGresult        *rc;
    ExecStatusType  est;

    if (!conn)
        return;

    rc = PQexec(conn, "BEGIN");
    if (!rc) {
        syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        transaction = 0;
    } else 
    if ((est = PQresultStatus(rc)) != PGRES_COMMAND_OK) {
        syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        transaction = 0;
    } else
        transaction = 1;
    PQclear(rc);
}

void commit(void *param)
{
    PGresult        *rc;
    ExecStatusType  est;

    if (!conn)
        return;

    if (transaction) {
        rc = PQexec(conn, "COMMIT");
        if (!rc) {
            syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        } else
        if ((est = PQresultStatus(rc)) != PGRES_COMMAND_OK) {
            syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        }
        PQclear(rc);
        transaction = 0;
    } else {
        /* fprintf(stderr, "No transaction in progress\n"); */
    }
}

void _close(void *param)
{
    if (!conn) return;

    commit(param);
    PQfinish(conn);

    if (dbhost) free(dbhost);
    if (dbuser) free(dbuser);
    if (dbpass) free(dbpass);
    if (dbname) free(dbname);
    if (dbheader) free(dbheader);
    if (dbflows) free(dbflows);
    if (dbextflows) free(dbextflows);
    if (dbproxy) free(dbproxy);
    if (ip_addr) free(ip_addr);
    if (intf) free(intf);

    syslog(LOG_DEBUG, "megabit: connection to database closed");
}

int isin(const int num, const int *arr, const int count)
{
    int	i = 0;

    for (i = 0; i < count; i++) {
        if (num == arr[i])
            return 1;
    }
    return 0;
}

void insertpkt(uint32_t address, void *packet, void *param)
{
    int             i, j;
    char            str[512];
    char            addr[128];
    struct tm       *stm;
    time_t          ttm;
    struct in_addr  in;
    short           version = ntohs(*((short *)packet));
    char            *tblname;
    uint32_t        dstaddr, srcaddr, nexthop;
    PGresult        *rc;
    ExecStatusType  est;
    unsigned int    uptime;
    unsigned int    luptime;
    struct flowhdr5_t   *hdr = (struct flowhdr5_t *)packet;
    struct flow5_t      *flow;

    if (!packet) return;

    if (!conn) return;

    /* only v5 netflow! */
    if (version != 5) return;

    /* verifying ip address */
    for (i = 0; i < addr_count; i++)
        if (ip_addr[i] == address) break;
    if (ip_addr[i] != address) {
        syslog(LOG_ERR, "megabit: packet from unknown ip address");
        return;
    }

    /* inserting flow header */
    in.s_addr = address;
    inet_ntop(AF_INET, &in, addr, sizeof(addr));

    if (ltime) 
        ttm = time(NULL);
    else {
        uptime = htonl((unsigned int)hdr->uptime);
        luptime = htonl((unsigned int)flow->luptime);

        if (uptime > luptime) {
            ttm = htonl(hdr->secs) - (uptime - luptime)/1000;
        } else {
            ttm = htonl(hdr->secs) + (luptime - uptime)/1000;
        }
    }
    stm = localtime(&ttm);

    /* insert flow header */
    if (savehdr && dbheader) {
        snprintf(str, sizeof(str), "INSERT INTO %s "
            "(address, version, count, uptime, secs, nsecs, "
            " flowseq, engine, eng_id, sampling, time) "
            "VALUES('%s', %d, %d, %lu, %lu, %lu, %lu, %d, %d, %d, "
            "'%04d-%02d-%02d %02d:%02d:%02d')",
            dbheader, addr, version, ntohs(hdr->count),
            (unsigned long)ntohl(hdr->uptime), (unsigned long)ntohl(hdr->secs),
            (unsigned long)ntohl(hdr->nsecs),
            (unsigned long)ntohl(hdr->flowseq),
            hdr->engine, hdr->eng_id, hdr->sampling,
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec);

        if (!(rc = PQexec(conn, str))) {
            syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        } else
        if ((est = PQresultStatus(rc)) != PGRES_COMMAND_OK) {
            syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        }
        PQclear(rc);
    }

    /* inserting flows */
    for (i = 0; i < ntohs(hdr->count); i++) {
        flow = packet + sizeof(struct flowhdr5_t) + sizeof(struct flow5_t)*i;

        dstaddr = *((uint32_t *)&flow->dstaddr);
        srcaddr = *((uint32_t *)&flow->srcaddr);
        nexthop = *((uint32_t *)&flow->nexthop);

        /* if nexthop is our netflow exporter - skip flow */
        for (j = 0; j < addr_count; j++) {
            if (ip_addr[j] == nexthop)
                break;
        }
        if (ip_addr[j] == nexthop)
            continue;

		/* determine table name */
        if (address == border &&
            ( isin(htons(flow->input_if), intf, intf_count) || 
            isin(htons(flow->output_if), intf, intf_count) )
        ) {
            tblname = dbextflows;
        } else 
        if ((dstaddr == proxy && htons(flow->dstport) == proxyport) ||
            (srcaddr == proxy && htons(flow->srcport) == proxyport)
        ) {
            tblname = dbproxy;
        } else {
            tblname = dbflows;
        }

        /* form sql query */
        snprintf(str, sizeof(str), "INSERT INTO %s "
            "(address, version, number, srcaddr, dstaddr, "
            "nexthop, input_if, output_if, dpkts, doctets, "
            "fuptime, luptime, srcport, dstport, resv1, "
            "flags, prot, tos, src_as, dst_as, srcmask, "
            "dstmask, resv2, time) "
            "VALUES('%s', %d, %lu, '%u.%u.%u.%u', '%u.%u.%u.%u', "
            "'%u.%u.%u.%u', %u, %u, %lu, %lu, %lu, %lu, %u, %u, "
            "%u, %u, %u, %u, %u, %u, %u, %u, %u, "
            "'%04d-%02d-%02d %02d:%02d:%02d')",
            tblname, addr, version, (unsigned long)(i+htonl(hdr->flowseq)),
            flow->srcaddr[0], flow->srcaddr[1],
            flow->srcaddr[2], flow->srcaddr[3],
            flow->dstaddr[0], flow->dstaddr[1],
            flow->dstaddr[2], flow->dstaddr[3],
            flow->nexthop[0], flow->nexthop[1],
            flow->nexthop[2], flow->nexthop[3],
            htons(flow->input_if), htons(flow->output_if),
            (unsigned long)htonl(flow->dpkts),
            (unsigned long)htonl(flow->doctets),
            (unsigned long)htonl(flow->fuptime),
            (unsigned long)htonl(flow->luptime),
            htons(flow->srcport), htons(flow->dstport),
            (unsigned char)flow->resv1, (unsigned char)flow->flags, 
            (unsigned char)flow->prot, (unsigned char)flow->tos, 
            htons(flow->src_as), htons(flow->dst_as),
            flow->srcmask, flow->dstmask, htons(flow->resv2),
            stm->tm_year+1900, stm->tm_mon+1, stm->tm_mday,
            stm->tm_hour, stm->tm_min, stm->tm_sec);
        if (!(rc = PQexec(conn, str))) {
            syslog(LOG_ERR, "megabit: %s", PQerrorMessage(conn));
        } else
        if ((est = PQresultStatus(rc)) != PGRES_COMMAND_OK) {
            fprintf(stderr, "megabit: %s in %s", PQerrorMessage(conn), str);
        }
        PQclear(rc);
    }
}
