/**************************************************************************
 * Copyright (c) 2009, Vyacheslav Yurkov <uvv.mail@gmail.com>             *
 * All rights reserved.                                                   *
 *                                                                        *
 * Redistribution and use in source and binary forms, with or without     *
 * modification, are permitted provided that the following conditions     *
 * are met:                                                               *
 *                                                                        *
 * 1. Redistributions of source code must retain the above copyright      *
 *    notice, this list of conditions and the following disclaimer.       *
 * 2. Redistributions in binary form must reproduce the above copyright   *
 *    notice, this list of conditions and the following disclaimer in the *
 *    documentation and/or other materials provided with the distribution.*
 * 3. Neither the name of the Vyacheslav Yurkov nor the names of his      *
 *    contributors may be used to endorse or promote products derived     *
 *    from this software without specific prior written permission.       *
 *                                                                        *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS    *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT      *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR  *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT   *
 * AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT       *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,  *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY  *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT    *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE  *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.   *
 **************************************************************************/

#include <sys/types.h>
#include <sys/socket.h>
#include <syslog.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <libpq-fe.h>

#include <libnnfc.h>

#define STMT_FLOW1HDR_T_INSERT "flow1hdr_t_insert"
#define STMT_FLOW5HDR_T_INSERT "flow5hdr_t_insert"
#define STMT_FLOW7HDR_T_INSERT "flow7hdr_t_insert"
#define STMT_FLOW1_T_INSERT "flow1_t_insert"
#define STMT_FLOW5_T_INSERT "flow5_t_insert"
#define STMT_FLOW7_T_INSERT "flow7_t_insert"

#define STMT_FLOW1HDR_T_NPARAMS 6
#define STMT_FLOW5HDR_T_NPARAMS 10
#define STMT_FLOW7HDR_T_NPARAMS 8
#define STMT_FLOW1_T_NPARAMS 22
#define STMT_FLOW5_T_NPARAMS 18
#define STMT_FLOW7_T_NPARAMS 22

typedef struct
{
    char address[19];
    uint32_t version;
    char number[11];
    char srcaddr[19];
    char dstaddr[19];
    char nexthop[19];
    uint32_t input_if;
    uint32_t output_if;
    char doctets[11];
    char dpkts[11];
    char fuptime[11];
    char luptime[11];
    uint32_t srcport;
    uint32_t dstport;
    uint32_t resv1;
    uint16_t flags;
    uint16_t prot;
    uint16_t tos;
    uint16_t mode;
    uint32_t src_as;
    uint32_t dst_as;
    uint16_t tcp_retx_cnt;
    uint16_t tcp_retx_sec;
    uint16_t tcp_misseq_cnt;
    char router[19];
    uint32_t resv2;
} flow_for_pg;

typedef struct
{
    char address[19];
    uint32_t version;
    uint32_t count;
    char uptime[11];
    char secs[11];
    char nsecs[11];
    char flowseq[11];
    uint16_t engine;
    uint16_t eng_id;
    uint32_t sampling;
    char resv[11];
    char time_id[20];
} header_for_pg;

typedef struct
{
    PGconn   *conn;
    char            config[MAXPATHLEN];
    /*
     *  Parameters for this module:
     *  localtime = store local or remote time?
     *  dbhost - postgres server
     *  dbport - postgres server's port (usually 5432)
     *  dbuser - postgres user's name
     *  dbpass - postgres user's password
     *  dbname - postgres database
     *  dbheader - postgres table for netflow headers
     *  dbflows - postgres table for netflow packets
     */

    int     ltime;
    char    *dbhost;
    int     dbport;
    char    *dbuser;
    char    *dbpass;
    char    *dbname;
    char    *dbheader;
    char    *dbflows;
    int     nonblock;
    char flow1hdr_t_insert[1024];
    char flow5hdr_t_insert[1024];
    char flow7hdr_t_insert[1024];
    char flow1_t_insert[1024];
    char flow5_t_insert[1024];
    char flow7_t_insert[1024];
    const char *flow1hdr_t_values[STMT_FLOW1HDR_T_NPARAMS];
    const char *flow5hdr_t_values[STMT_FLOW5HDR_T_NPARAMS];
    const char *flow7hdr_t_values[STMT_FLOW7HDR_T_NPARAMS];
    const char *flow1_t_values[STMT_FLOW1_T_NPARAMS];
    const char *flow5_t_values[STMT_FLOW5_T_NPARAMS];
    const char *flow7_t_values[STMT_FLOW7_T_NPARAMS];
    flow_for_pg flow_t_packet;
    header_for_pg flowhdr_t_packet;
    int flow1hdr_t_paramformats[STMT_FLOW1HDR_T_NPARAMS];
    int flow5hdr_t_paramformats[STMT_FLOW5HDR_T_NPARAMS];
    int flow7hdr_t_paramformats[STMT_FLOW7HDR_T_NPARAMS];
    int flow1_t_paramformats[STMT_FLOW1_T_NPARAMS];
    int flow5_t_paramformats[STMT_FLOW5_T_NPARAMS];
    int flow7_t_paramformats[STMT_FLOW7_T_NPARAMS];
    int flow1hdr_t_paramlengths[STMT_FLOW1HDR_T_NPARAMS];
    int flow5hdr_t_paramlengths[STMT_FLOW5HDR_T_NPARAMS];
    int flow7hdr_t_paramlengths[STMT_FLOW7HDR_T_NPARAMS];
    int flow1_t_paramlengths[STMT_FLOW1_T_NPARAMS];
    int flow5_t_paramlengths[STMT_FLOW5_T_NPARAMS];
    int flow7_t_paramlengths[STMT_FLOW7_T_NPARAMS];
} thread_context;
